"""Support for checking different CLI input types"""

# -- Imports ------------------------------------------------------------------

from argparse import ArgumentTypeError
from math import inf
from re import compile as re_compile

from netaddr import AddrFormatError, EUI

# -- Functions ----------------------------------------------------------------


def base64_mac_address(name):
    """Check if the given text represents a Base64 encoded MAC address

    Returns:

        The given text on success

    Raises:

        An argument type error in case the given value does not represent a MAC
        address

    Examples:

        Parse valid Base64 encoded MAC address

        >>> base64_mac_address("CGvXAd6B")
        'CGvXAd6B'

        Parse invalid Base64 encoded MAC address

        >>> base64_mac_address("CGvXAd") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “CGvXAd” is not a Base64 encoded MAC
                                    address

    """

    base64_regex = re_compile("[A-Za-z0-9/+]{8}$")
    if base64_regex.match(name):
        return name
    raise ArgumentTypeError(f"“{name}” is not a Base64 encoded MAC address")


def byte_value(value):
    """Check if the given string represents a byte value

    Returns:

        An integer representing the given value on success

    Raises:

        An argument type error in case the given value does not represent a
        (positive) byte value

    Examples:

        Parse valid byte values

        >>> byte_value("0xa")
        10
        >>> byte_value("137")
        137

        Parse incorrect byte value

        >>> byte_value("256")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “256” is not a valid byte value

    """

    try:
        number = int(value, base=0)
        if number < 0 or number > 255:
            raise ValueError()
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{value}” is not a valid byte value"
        ) from error


def channel_number(value: str):
    """Check if the given string represents a valid channel number (0 – 255)

    Raises:

        An argument type error in case the given value does not represent a
        channel number

    Returns:

        An integer representing the given channel number on success

    Examples:

        Parse correct channel numbers

        >>> channel_number("1")
        1
        >>> channel_number("123")
        123
        >>> channel_number("0")
        0
        >>> channel_number("255")
        255

        Parsing an incorrect negative channel number will fail

        >>> channel_number("-1")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “-1” is not a valid channel number

        Parse an incorrectly large channel number will fail

        >>> channel_number("256")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “256” is not a valid channel number

    """

    try:
        number: int | None = None
        number = int(value)
        if number < 0 or number > 255:
            raise ValueError()
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{number if number is not None else value}” is not a valid "
            "channel number"
        ) from error


def mac_address(address: str) -> EUI:
    """Check if the given text represents a MAC address

    Returns:

        The parsed MAC address

    Raises:

        An argument type error in case the given text does not store a MAC
        address of the form

        - ``xx:xx:xx:xx:xx:xx`` or
        - ``xx-xx-xx-xx-xx-xx``

        where `x` represents a hexadecimal number.

    Examples:

        Parse a correct MAC address

        >>> mac_address("08:6b:d7:01:de:81")
        EUI('08-6B-D7-01-DE-81')

        Parsing an incorrect MAC address will fails

        >>> mac_address("08:6b:d7:01:de:666") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “08:6b:d7:01:de:666” is not a valid MAC
                                    address

    """

    try:
        eui = EUI(address)
    except AddrFormatError as error:
        raise ArgumentTypeError(
            f"“{address}” is not a valid MAC address"
        ) from error

    return eui


def measurement_time(value: str) -> float:
    """Check if the given number is valid measurement time

    0 will be interpreted as infinite measurement runtime

    Returns:

        A float value representing the measurement time on success

    Raises:

        ArgumentTypeError:
             If the given text is not a valid measurement time value


    Examples:

        Parse correct measurement times

        >>> measurement_time("0")
        inf

        >>> measurement_time("0.1")
        0.1

        >>> measurement_time("12.34")
        12.34

        Parsing a negative measurement time fails

        >>> measurement_time("-1")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “-1” is not a valid measurement time

        Parsing an incorrect measurement time fails

        >>> measurement_time("something")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “something” is not a valid measurement time

    """

    try:
        number = float(value)
        if number < 0:
            raise ValueError()
        if number == 0:
            return inf
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{value}” is not a valid measurement time"
        ) from error


def non_infinite_measurement_time(value: str) -> float:
    """Check if the given number is valid (non-infinite) measurement time

    0 will be interpreted as infinite measurement runtime

    Returns:

        A float value representing the measurement time on success

    Raises:

        ArgumentTypeError:
             If the given text is not a valid measurement time value

    Examples:

        Parse correct measurement time

        >>> non_infinite_measurement_time("0.1")
        0.1

        Parsing an “infinite” measurement time fails

        >>> non_infinite_measurement_time("0") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “0” is not a valid positive measurement
                                    time

    """

    runtime = measurement_time(value)
    if runtime == inf:
        raise ArgumentTypeError(
            f"“{value}” is not a valid positive measurement time"
        )

    return runtime


def sensor_node_number(value: str) -> int:
    """Check if the given number is valid Bluetooth node number

    Returns:

        An integer representing the given channel number on success

    Raises:

        ArgumentTypeError:
             If the given text is not a valid sensor node number

    Examples:

        Parse correct sensor node numbers

        >>> sensor_node_number("0")
        0

        >>> sensor_node_number("123")
        123

        Parsing an incorrect negative sensor node number fails

        >>> sensor_node_number("-1")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “-1” is not a valid Bluetooth node number

        Parsing an incorrect sensor node number fails

        >>> sensor_node_number("hello") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “hello” is not a valid Bluetooth node
                                    number

    """

    try:
        number = int(value)
        if number < 0:
            raise ValueError()
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{value}” is not a valid Bluetooth node number"
        ) from error


def node_name(name: str) -> str:
    """Check if the given text is a valid node name

    Returns:

        The given name on success

    Raises:

        An argument type error in case the given text does not store a valid
        node name. This is the case, if the name

        - is longer than 8 characters or
        - contains non-ASCII data.

    Examples:

        Parse a correct node name

        >>> node_name("Blubb")
        'Blubb'

        A node name is not allowed to contain non-ASCII data

        >>> node_name("Blübb")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “Blübb” is not a valid name

        A node name can not be longer than 8 characters

        >>> node_name("123456789") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “123456789” is too long to be a valid
                                    name

    """

    try:
        name.encode("ascii")
    except UnicodeEncodeError as error:
        raise ArgumentTypeError(f"“{name}” is not a valid name") from error

    if len(name) > 8:
        raise ArgumentTypeError(f"“{name}” is too long to be a valid name")

    return name



"""Command Line Parsing support"""

# -- Imports ------------------------------------------------------------------

from argparse import ArgumentParser

from icotronic.can.adc import ADCConfiguration
from icotronic.cmdline.types import (
    channel_number,
    node_name,
    mac_address,
    measurement_time,
    non_infinite_measurement_time,
    sensor_node_number,
)

# -- Functions ----------------------------------------------------------------


def add_identifier_arguments(parser: ArgumentParser) -> None:
    """Add node identifier arguments to given argument parser

    Args:

        parser:
            The parser which should include the node identifier arguments

    """

    identifier_arg_group = parser.add_argument_group(
        title="Sensor Node Identifier"
    )

    identifier_group = identifier_arg_group.add_mutually_exclusive_group(
        required=True
    )

    identifier_group.add_argument(
        "-n",
        "--name",
        dest="identifier",
        metavar="NAME",
        help="Name of sensor node",
        default="Test-STH",
        type=node_name,
    )
    identifier_group.add_argument(
        "-m",
        "--mac-address",
        dest="identifier",
        metavar="MAC_ADRESS",
        help="Bluetooth MAC address of sensor node",
        type=mac_address,
    )
    identifier_group.add_argument(
        "-d",
        "--number",
        type=sensor_node_number,
        dest="identifier",
        metavar="NUMBER",
        help="Node number of sensor node",
    )


def add_adc_arguments(parser: ArgumentParser) -> None:
    """Add ADC arguments to given argument parser

    Args:

        parser:
            The parser which should include the ADC arguments

    """

    adc_group = parser.add_argument_group(title="ADC")
    adc_group.add_argument(
        "-s",
        "--prescaler",
        type=int,
        choices=range(2, 128),
        metavar="2–127",
        default=2,
        required=False,
        help="Prescaler value",
    )
    adc_group.add_argument(
        "-a",
        "--acquisition",
        type=int,
        choices=sorted([2**number for number in range(9)] + [3]),
        default=8,
        required=False,
        help="Acquisition time value",
    )
    adc_group.add_argument(
        "-o",
        "--oversampling",
        type=int,
        choices=[2**number for number in range(13)],
        default=64,
        required=False,
        help="Oversampling rate value",
    )
    adc_group.add_argument(
        "-v",
        "--voltage-reference",
        choices=ADCConfiguration.REFERENCE_VOLTAGES,
        type=float,
        default=3.3,
        required=False,
        help="Reference voltage in V",
    )


def add_channel_arguments(group) -> None:
    """Add channel arguments to given argument parser group

    Args:

        group:
            The parser group to which the channel arguments should be added to

    """

    group.add_argument(
        "-1",
        "--first-channel",
        type=channel_number,
        default=1,
        const=1,
        nargs="?",
        help=(
            "sensor channel number for first measurement channel "
            "(1 - 255; 0 to disable)"
        ),
    )
    group.add_argument(
        "-2",
        "--second-channel",
        type=channel_number,
        default=0,
        const=2,
        nargs="?",
        help=(
            "sensor channel number for second measurement channel "
            "(1 - 255; 0 to disable)"
        ),
    )
    group.add_argument(
        "-3",
        "--third-channel",
        type=channel_number,
        default=0,
        const=3,
        nargs="?",
        help=(
            "sensor channel number for third measurement channel "
            "(1 - 255; 0 to disable)"
        ),
    )


def create_icon_parser() -> ArgumentParser:
    """Create command line parser for icon

    Returns:

        A parser for the CLI arguments of icon

    """

    parser = ArgumentParser(description="ICOtronic CLI tool")

    parser.add_argument(
        "--log",
        choices=("debug", "info", "warning", "error", "critical"),
        default="warning",
        required=False,
        help="minimum log level",
    )

    subparsers = parser.add_subparsers(
        required=True, title="Subcommands", dest="subcommand"
    )

    # ==========
    # = Config =
    # ==========

    subparsers.add_parser(
        "config", help="Open config file in default application"
    )

    # =============
    # = Data Loss =
    # =============

    dataloss_parser = subparsers.add_parser(
        "dataloss", help="Check data loss at different sample rates"
    )
    measurement_group = dataloss_parser.add_argument_group(title="Measurement")
    measurement_group.add_argument(
        "-t",
        "--time",
        type=non_infinite_measurement_time,
        help="Measurement time in seconds",
        default=10,
    )
    add_identifier_arguments(dataloss_parser)

    # ========
    # = List =
    # ========

    subparsers.add_parser("list", help="List sensor nodes")

    # ===========
    # = Measure =
    # ===========

    measurement_parser = subparsers.add_parser(
        "measure", help="Store measurement data"
    )

    measurement_group = measurement_parser.add_argument_group(
        title="Measurement"
    )

    measurement_group.add_argument(
        "-t",
        "--time",
        type=measurement_time,
        help="Measurement time in seconds (0 for infinite runtime)",
        default=10,
    )

    add_channel_arguments(measurement_group)
    add_identifier_arguments(measurement_parser)
    add_adc_arguments(measurement_parser)

    # ==========
    # = Rename =
    # ==========

    rename_parser = subparsers.add_parser(
        "rename", help="Rename a sensor node"
    )
    add_identifier_arguments(rename_parser)
    rename_parser.add_argument(
        "name",
        type=str,
        help="New name of sensor node",
        nargs="?",
        default="Test-STH",
    )

    # =======
    # = STU =
    # =======

    stu_parser = subparsers.add_parser(
        "stu", help="Execute commands related to stationary receiver unit"
    )

    stu_subparsers = stu_parser.add_subparsers(
        required=True, title="Subcommands", dest="stu_subcommand"
    )
    stu_subparsers.add_parser("mac", help="Show Bluetooth MAC address")
    stu_subparsers.add_parser("reset", help="Reset STU")

    return parser


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for reading EEPROM status"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

# -- Class --------------------------------------------------------------------


class EEPROMStatus:
    """This class represents an EEPROM status byte

    Args:

        status:
            The value of the status byte

    Examples:

        Create some EEPROM status objects from scratch

        >>> EEPROMStatus(0xca)
        Locked (0xca)

        >>> EEPROMStatus(0)
        Uninitialized (0x00)

        >>> EEPROMStatus('Initialized')
        Initialized (0xac)

        >>> EEPROMStatus('Uninitialized')
        Uninitialized (0x00)

        >>> EEPROMStatus('Locked').value == 0xca
        True

        >>> EEPROMStatus(EEPROMStatus('Locked'))
        Locked (0xca)

        Initializing the EEPROM status with an incorrect value will fail

        >>> EEPROMStatus('Something')
        Traceback (most recent call last):
           ...
        ValueError: Unknown EEPROM status “Something”

    """

    def __init__(self, status: int | str | EEPROMStatus) -> None:

        if isinstance(status, str):
            if status == "Initialized":
                self.value = 0xAC
                return

            if status == "Locked":
                self.value = 0xCA
                return

            if status == "Uninitialized":
                self.value = 0
                return

            raise ValueError(f"Unknown EEPROM status “{status}”")

        if isinstance(status, EEPROMStatus):
            self.value = status.value
            return

        self.value = status

    def __eq__(self, other: object) -> bool:
        """Compare this EEPROM status to another object

        Args:

            other:

                The other object the status should be compared to

        Returns:

            - ``True``, if the given object is a EEPROM status and it has the
              same value as this state

            - ``False``, otherwise

        Examples:

            Compare some example EEPROM status objects

            >>> initialized = EEPROMStatus('Initialized')
            >>> locked = EEPROMStatus('Locked')

            >>> initialized == EEPROMStatus('Initialized')
            True
            >>> locked == EEPROMStatus(0xCA)
            True
            >>> locked == initialized
            False


        """

        if isinstance(other, EEPROMStatus):
            return self.value == other.value

        return False

    def __repr__(self) -> str:
        """Return the string representation of the status byte

        Returns:

            A string that describes the current value of the status byte

        Examples:

            Get the string representation of various status bytes

            >>> EEPROMStatus(0xac)
            Initialized (0xac)

            >>> EEPROMStatus(0x13)
            Uninitialized (0x13)

            >>> EEPROMStatus(0xCA)
            Locked (0xca)

        """

        value = self.value

        description = (
            "Initialized"
            if self.is_initialized()
            else ("Locked" if self.is_locked() else "Uninitialized")
        )

        return f"{description} (0x{value:02x})"

    def is_locked(self) -> bool:
        """Check if the EEPROM is locked

        Returns:

            ``True`` if the status byte represents a locked EEPROM or `False`
            otherwise

        Examples:

            Check the lock status of various EEPROM status bytes

            >>> EEPROMStatus(0xca).is_locked()
            True

            >>> EEPROMStatus(0xac).is_locked()
            False

            >>> EEPROMStatus(0x2).is_locked()
            False

        """

        return self.value == 0xCA

    def is_initialized(self) -> bool:
        """Check if the EEPROM is initialized

        Returns:

        ``True`` if the status byte represents an initialized EEPROM or
        ``False`` otherwise

        Examples:

            Check the initialization status of some EEPROM status bytes

            >>> EEPROMStatus(0xca).is_initialized()
            False

            >>> EEPROMStatus(0xac).is_initialized()
            True

            >>> EEPROMStatus(0x2).is_initialized()
            False

        """

        return self.value == 0xAC


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Show concurrent capabilities of ICOtronic system"""

# -- Imports ------------------------------------------------------------------

from asyncio import Event, run, TaskGroup
from contextlib import asynccontextmanager
from typing import AsyncIterator

from netaddr import EUI

from icotronic.can import Connection, SensorNode, StreamingConfiguration
from icotronic.can.adc import ADCConfiguration

# -- Functions ----------------------------------------------------------------


@asynccontextmanager
async def sensor_node_connection(
    identifier: int | str | EUI,
) -> AsyncIterator[SensorNode]:
    """Create a connection to a sensor node

    Args:

        The identifier of the sensor node

    Returns:

        A asynchronous context manager for accessing the sensor node

    """

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier) as sensor_node:
            yield sensor_node


async def stream_and_read_adc():
    """Read ADC configuration, while streaming is active"""

    async with sensor_node_connection("Test-STH") as sensor_node:

        async def stream_data(started_streaming: Event) -> None:
            """Enable streaming and read streaming data until cancelled"""

            async with sensor_node.open_data_stream(
                StreamingConfiguration(first=True)
            ) as stream:
                async for _ in stream:
                    if not started_streaming.is_set():
                        started_streaming.set()

        async def read_adc() -> ADCConfiguration:
            """Read ADC configuration"""

            return await sensor_node.get_adc_configuration()

        started_streaming = Event()

        async with TaskGroup() as task_group:
            stream_data_task = task_group.create_task(
                stream_data(started_streaming)
            )
            await started_streaming.wait()  # Wait until streaming is active
            read_adc_task = task_group.create_task(read_adc())
            adc_configuration = await read_adc_task
            print(adc_configuration)
            stream_data_task.cancel()


# -- Main ---------------------------------------------------------------------


if __name__ == "__main__":
    run(stream_and_read_adc())
